"""The package mecab defines Japanese spacy.Language with Mecab tokenizer."""
import shutil
from collections import namedtuple
from pathlib import Path
from shutil import copytree
from typing import List, Optional, Type

from spacy.compat import copy_reg
from spacy.language import Language
from spacy.tokens import Doc, Token

from camphr.consts import KEY_FSTRING
from camphr.lang.stop_words import STOP_WORDS
from camphr.utils import RE_URL, SerializationMixin

ShortUnitWord = namedtuple(
    "ShortUnitWord", ["surface", "lemma", "pos", "space", "fstring"]
)


class Tokenizer(SerializationMixin):
    USERDIC = "user.dic"  # used when saving
    ASSETS = "assets"  # used when saving
    key_fstring = KEY_FSTRING

    @classmethod
    def install_extensions(cls):
        Token.set_extension(cls.key_fstring, default=None, force=True)

    def __init__(
        self,
        cls: Type["Defaults"],
        nlp: Optional[Language] = None,
        dicdir: Optional[str] = None,
        userdic: Optional[str] = None,
        assets: Optional[str] = None,
    ):
        """

        Args:
            dicdir: Mecab dictionary path. If `None`, use system configuration (~/.mecabrc or /usr/local/etc/mecabrc).
            userdic: Mecab user dictionary path. If `None`, use system configuration (~/.mecabrc or /usr/local/etc/mecabrc).
            assets: Other assets path saved with tokenizer. e.g. userdic definition csv path
        """
        self.vocab = nlp.vocab if nlp is not None else cls.create_vocab(nlp)
        self.tokenizer = self.get_mecab(dicdir=dicdir, userdic=userdic)
        self.assets = assets

    def __call__(self, text: str) -> Doc:
        dtokens = self.detailed_tokens(text)
        words = [x.surface for x in dtokens]
        spaces = [x.space for x in dtokens]
        doc = Doc(self.vocab, words=words, spaces=spaces)
        for token, dtoken in zip(doc, dtokens):
            token.tag_ = dtoken.pos
            token.lemma_ = dtoken.lemma
            token._.set(self.key_fstring, dtoken.fstring)

        with doc.retokenize() as retokenizer:
            for match in RE_URL.finditer(doc.text):
                span = doc.char_span(*match.span())
                if span:
                    retokenizer.merge(span)
        doc.is_tagged = True
        return doc

    def detailed_tokens(self, text: str) -> List[ShortUnitWord]:
        """Tokenize text with Mecab and format the outputs for further processing"""
        node = self.tokenizer.parseToNode(text)
        node = node.next
        words: List[ShortUnitWord] = []
        while node.posid != 0:
            parts = node.feature.split(",")
            pos = ",".join(parts[0:4])
            surface = node.surface
            base = surface
            if len(parts) > 6:
                base = parts[6]
            nextnode = node.next
            if nextnode.length != nextnode.rlength:
                # next node contains space, so attach it to this node.
                words.append(ShortUnitWord(surface, base, pos, True, node.feature))
            elif nextnode.surface == "\u3000":
                # next node is full space, so attatch it to this node and skip the nextnode.
                words.append(ShortUnitWord(surface, base, pos, True, node.feature))
                nextnode = nextnode.next
            else:
                words.append(ShortUnitWord(surface, base, pos, False, node.feature))
            node = nextnode
        return words

    def get_mecab(self, dicdir: Optional[str] = None, userdic: Optional[str] = None):
        """Create `MeCab.Tagger` instance"""
        import MeCab

        opt = ""
        if userdic:
            opt += f"-u {userdic} "
        if dicdir:
            opt += f"-d {dicdir} "
        self.userdic = userdic
        self.dicdir = dicdir
        tokenizer = MeCab.Tagger(opt.strip())
        tokenizer.parseToNode("")  # see https://github.com/explosion/spaCy/issues/2901
        return tokenizer

    def to_disk(self, path: Path, **kwargs):
        path.mkdir(exist_ok=True)
        if self.userdic:
            shutil.copy(self.userdic, path / self.USERDIC)
        if self.assets:
            copytree(self.assets, path / self.ASSETS)

    def from_disk(self, path: Path, **kwargs):
        """TODO: is userdic portable?"""
        userdic = (path / self.USERDIC).absolute()
        if userdic.exists():
            self.userdic = str(userdic)
        self.tokenizer = self.get_mecab(userdic=self.userdic)

        assets = (path / self.ASSETS).absolute()
        if assets.exists():
            self.assets = str(assets)
        return self


# for pickling. see https://spacy.io/usage/adding-languages
class Defaults(Language.Defaults):  # type: ignore
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)
    stop_words = STOP_WORDS
    writing_system = {"direction": "ltr", "has_case": False, "has_letters": False}

    @classmethod
    def create_tokenizer(
        cls, nlp=None, dicdir: str = None, userdic: str = None, assets: str = None
    ):
        return Tokenizer(cls, nlp, dicdir=dicdir, userdic=userdic, assets=assets)


class Japanese(Language):
    lang = "ja_mecab"
    Defaults = Defaults

    def make_doc(self, text: str) -> Doc:
        return self.tokenizer(text)


# avoid pickling problem (see https://github.com/explosion/spaCy/issues/3191)
def pickle_japanese(instance):
    return Japanese, tuple()


copy_reg.pickle(Japanese, pickle_japanese)
Language.factories[Japanese.lang] = Japanese

Tokenizer.install_extensions()


# for lazy loading. see https://spacy.io/usage/adding-languages
__all__ = ["Japanese"]



import re
from typing import Dict, List, Optional, Pattern, Union

import spacy
from spacy.tokens import Doc, Span
from spacy.util import filter_spans

from camphr.utils import SerializationMixin, get_doc_char_spans_list, merge_spans


@spacy.component(
    "multiple_regex_ruler", assigns=["doc.ents", "token.ent_type"], retokenizes=True
)
class MultipleRegexRuler(SerializationMixin):
    serialization_fields = ["patterns", "destructive", "merge"]
    _DEFAULT_LABEL = "matched"

    def __init__(
        self,
        patterns: Optional[Dict[str, Union[Pattern, str]]] = None,
        destructive: bool = False,
        merge: bool = False,
    ):
        self.patterns = patterns or {}
        self.destructive = destructive
        self.merge = merge

    def require_model(self):
        assert self.patterns

    @property
    def labels(self) -> List[str]:
        return list(self.patterns)  # type: ignore

    def __call__(self, doc: Doc) -> Doc:
        self.require_model()
        for label, pattern in self.patterns.items():  # type: ignore
            doc = self._proc(doc, pattern, label)
        return doc

    def _proc(self, doc: Doc, pattern: Union[Pattern, str], label: str) -> Doc:
        spans = self.get_spans(doc, pattern, label or self._DEFAULT_LABEL)
        doc.ents = filter_spans(tuple(spans) + doc.ents)  # type: ignore
        # TODO: https://github.com/python/mypy/issues/3004
        if self.merge:
            merge_spans(doc, spans)
        return doc

    def get_spans(
        self, doc: Doc, pattern: Union[Pattern, str], label: str
    ) -> List[Span]:
        spans_ij = [m.span() for m in re.finditer(pattern, doc.text)]  # type: ignore
        return get_doc_char_spans_list(
            doc, spans_ij, destructive=self.destructive, label=label
        )

    @classmethod
    def from_nlp(cls, *args, **kwargs) -> "MultipleRegexRuler":
        return cls()


@spacy.component(
    "regex_ruler", assigns=["doc.ents", "token.ent_type"], retokenizes=True
)
class RegexRuler(MultipleRegexRuler):
    def __init__(
        self,
        pattern: Union[Pattern, str] = "",
        label: str = "",
        destructive: bool = False,
        merge: bool = False,
        name: str = "",
    ):
        self.patterns = {label: pattern}
        self.destructive = destructive
        self.merge = merge
        if name:
            self.name = name
        else:
            self.name = "regex_ruler_" + label



from typing import NamedTuple


class KnpUserKeyType(NamedTuple):
    element: str  # pyknp.Bunsetsu, Tag, Morpheme
    spans: str  # span list containing Bunsetsu, Tag, Morpheme correspondings.
    list_: str  # list containing knp elements
    parent: str
    children: str
    tag: str
    bunsetsu: str


class KnpUserKeys(NamedTuple):
    tag: KnpUserKeyType
    bunsetsu: KnpUserKeyType
    morph: KnpUserKeyType


KNP_USER_KEYS = KnpUserKeys(
    *[
        KnpUserKeyType(
            *["knp_" + comp + f"_{type_}" for type_ in KnpUserKeyType._fields]
        )
        for comp in KnpUserKeys._fields
    ]
)



"""The package juman defines Japanese spacy.Language with JUMAN tokenizer."""
import itertools
from collections import namedtuple
from typing import Any, Callable, Dict, Iterator, List, Optional, Type

from spacy.compat import copy_reg
from spacy.language import Language
from spacy.tokens import Doc, Token

from camphr.consts import JUMAN_LINES, KEY_FSTRING
from camphr.lang.stop_words import STOP_WORDS
from camphr.utils import SerializationMixin, get_juman_command

from .tag_map import TAG_MAP

ShortUnitWord = namedtuple(
    "ShortUnitWord", ["surface", "lemma", "pos", "fstring", "space"]
)
_REPLACE_STRINGS = {"\t": "　", "\r": "", "\n": "　"}


def han_to_zen_normalize(text):
    try:
        import mojimoji
    except ImportError:
        raise ValueError("juman or knp Language requires mojimoji.")
    text = mojimoji.han_to_zen(text)
    for k, v in _REPLACE_STRINGS.items():
        text = text.replace(k, v)
    return text


class Tokenizer(SerializationMixin):
    """Juman tokenizer

    Note:
        `spacy.Token._.fstring` is set. The Juman's output is stored into it during tokenizing.
    """

    serialization_fields = ["preprocessor", "juman_kwargs"]
    key_fstring = KEY_FSTRING

    @classmethod
    def install_extensions(cls):
        """See https://github.com/explosion/spacy-pytorch-transformers#extension-attributes."""
        Token.set_extension(cls.key_fstring, default=None, force=True)

    def __init__(
        self,
        cls: Type["Defaults"],
        nlp: Optional[Language] = None,
        juman_kwargs: Optional[Dict[str, str]] = None,
        preprocessor: Optional[Callable[[str], str]] = han_to_zen_normalize,
    ):
        """

        Args:
            juman_kwargs: passed to `pyknp.Juman.__init__`
            preprocessor: applied to text before tokenizing. `mojimoji.han_to_zen` is often used.
        """
        from pyknp import Juman

        juman_kwargs = juman_kwargs or {}
        default_command = get_juman_command()
        assert default_command
        juman_kwargs.setdefault("command", default_command)

        self.vocab = nlp.vocab if nlp is not None else cls.create_vocab(nlp)
        self.tokenizer = Juman(**juman_kwargs) if juman_kwargs else Juman()
        self.juman_kwargs = juman_kwargs
        self.preprocessor = preprocessor

    def reset_tokenizer(self):
        from pyknp import Juman

        self.tokenizer = Juman(**self.juman_kwargs) if self.juman_kwargs else Juman()

    def __call__(self, text: str) -> Doc:
        """Make doc from text. Juman's `fstring` is stored in `Token._.fstring`"""
        if self.preprocessor:
            text = self.preprocessor(text)
        juman_lines = self._juman_string(text)
        dtokens = self._detailed_tokens(juman_lines)
        doc = self._dtokens_to_doc(dtokens)
        doc.user_data[JUMAN_LINES] = juman_lines
        return doc

    def _juman_string(self, text: str) -> str:
        try:
            texts = _split_text_for_juman(text)
            lines: str = "".join(
                itertools.chain.from_iterable(
                    self.tokenizer.juman_lines(text) for text in texts
                )
            )
        except BrokenPipeError:
            # Juman is sometimes broken due to its subprocess management.
            self.reset_tokenizer()
            lines = self.tokenizer.juman_lines(text)
        return lines

    def _dtokens_to_doc(self, dtokens: List[ShortUnitWord]) -> Doc:
        words = [x.surface for x in dtokens]
        spaces = [x.space for x in dtokens]
        doc = Doc(self.vocab, words=words, spaces=spaces)
        for token, dtoken in zip(doc, dtokens):
            token.tag_ = dtoken.pos
            token.lemma_ = dtoken.lemma
            token._.set(self.key_fstring, dtoken.fstring)
        doc.is_tagged = True
        return doc

    def _detailed_tokens(self, juman_lines: str) -> List[ShortUnitWord]:
        """Tokenize text with Juman and format the outputs for further processing"""
        from pyknp import MList

        ml = MList(juman_lines).mrph_list()
        words: List[ShortUnitWord] = []
        for m in ml:
            surface = m.midasi
            pos = m.hinsi + "," + m.bunrui
            lemma = m.genkei or surface
            words.append(ShortUnitWord(surface, lemma, pos, m.fstring, False))
        return words


_SEPS = ["。", ".", "．"]


def _split_text_for_juman(text: str) -> Iterator[str]:
    """Juman denies long text (maybe >4096 bytes) so split text"""
    n = 1000
    if len(text) == 0:
        return
    if len(text) < n:
        yield text
        return
    for sep in _SEPS:
        if sep in text:
            i = text.index(sep)
            head, tail = text[: i + 1], text[i + 1 :]
            if len(head) < n:
                yield from _split_text_for_juman(head)
                yield from _split_text_for_juman(tail)
                return
    # If any separator is not found in text, split roughly
    yield text[:n]
    yield from _split_text_for_juman(text[n:])


# for pickling. see https://spacy.io/usage/adding-languages
class Defaults(Language.Defaults):  # type: ignore
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)
    tag_map = TAG_MAP
    stop_words = STOP_WORDS
    writing_system = {"direction": "ltr", "has_case": False, "has_letters": False}

    @classmethod
    def create_tokenizer(
        cls,
        nlp=None,
        juman_kwargs: Optional[Dict[str, Any]] = None,
        preprocessor: Optional[Callable[[str], str]] = han_to_zen_normalize,
    ):
        return Tokenizer(cls, nlp, juman_kwargs=juman_kwargs, preprocessor=preprocessor)


class Japanese(Language):
    lang = "ja_juman"
    Defaults = Defaults

    def make_doc(self, text: str) -> Doc:
        return self.tokenizer(text)


# avoid pickling problem (see https://github.com/explosion/spaCy/issues/3191)
def pickle_japanese(instance):
    return Japanese, tuple()


copy_reg.pickle(Japanese, pickle_japanese)
Language.factories[Japanese.lang] = Japanese

# for lazy loading. see https://spacy.io/usage/adding-languages
__all__ = ["Japanese"]

Tokenizer.install_extensions()



CARDINAL = "CARDINAL"
DATE = "DATE"
EVENT = "EVENT"
FAC = "FAC"
GPE = "GPE"
LANGUAGE = "LANGUAGE"
LAW = "LAW"
LOC = "LOC"
MONEY = "MONEY"
NORP = "NORP"
ORDINAL = "ORDINAL"
ORG = "ORG"
PERCENT = "PERCENT"
PERSON = "PERSON"
PRODUCT = "PRODUCT"
QUANTITY = "QUANTITY"
TIME = "TIME"
WORK_OF_ART = "WORK_OF_ART"

ALL_LABELS = [
    CARDINAL,
    DATE,
    EVENT,
    FAC,
    GPE,
    LANGUAGE,
    LAW,
    LOC,
    MONEY,
    NORP,
    ORDINAL,
    ORG,
    PERCENT,
    PERSON,
    PRODUCT,
    QUANTITY,
    TIME,
    WORK_OF_ART,
]



# flake8: noqa
from .maskedlm import (
    BertForMaskedLM,
    BertForMaskedLMPreprocessor,
    add_maskedlm_pipe,
    remove_maskedlm_pipe,
)
from .model import TrfModel
from .ner import TrfForNamedEntityRecognition
from .seq_classification import (
    TrfForMultiLabelSequenceClassification,
    TrfForSequenceClassification,
)
from .tokenizer import TrfTokenizer



"""Defines transformers NER pipe"""
from typing import Dict, Iterable, Iterator, List, Sequence, Sized, cast

import spacy
import torch
import torch.nn as nn
import torch.nn.functional as F
import transformers
from spacy.gold import GoldParse, iob_to_biluo, spans_from_biluo_tags
from spacy.tokens import Doc

from camphr.pipelines.utils import (
    UNK,
    B,
    I,
    L,
    U,
    UserHooksMixin,
    beamsearch,
    construct_biluo_tag,
    correct_bio_tags,
    deconstruct_biluo_label,
)
from camphr.torch_utils import TorchPipe, add_loss_to_docs

from .auto import get_trf_config_cls
from .utils import (
    ATTRS,
    LABELS,
    EstimatorMixin,
    FromNLPMixinForTrfTask,
    LabelsMixin,
    SerializationMixinForTrfTask,
    TrfModelForTaskBase,
    get_dropout,
    get_last_hidden_state_from_docs,
)

NUM_LABELS = "num_labels"


class TrfTokenClassifier(TrfModelForTaskBase):
    """Head layer for classification task"""

    def __init__(self, config: transformers.PretrainedConfig):
        super().__init__(config)
        self.num_labels = config.num_labels
        dropout = get_dropout(config)
        hidden_size = config.hidden_size
        self.dropout = nn.Dropout(dropout)
        self.classifier = nn.Linear(hidden_size, self.num_labels)
        self.loss_fct = nn.CrossEntropyLoss()

    def forward(  # type:ignore
        self, x: torch.Tensor, mask: torch.Tensor = None, labels=None
    ) -> torch.Tensor:

        x = self.dropout(x)
        logits = self.classifier(x)
        return logits


class TrfForTokenClassificationBase(
    EstimatorMixin[torch.Tensor],
    LabelsMixin,
    FromNLPMixinForTrfTask,
    UserHooksMixin,
    SerializationMixinForTrfTask,
    TorchPipe,
):
    """Base class for token classification task (e.g. Named entity recognition).

    Requires `TrfModel` before this model in the pipeline to use this model.
    """

    model_cls = TrfTokenClassifier

    @classmethod
    def Model(cls, trf_name_or_path: str, **cfg) -> TrfTokenClassifier:
        config = get_trf_config_cls(trf_name_or_path).from_pretrained(trf_name_or_path)
        setattr(config, NUM_LABELS, len(cfg[LABELS]))
        return TrfTokenClassifier(config)

    def proc_model(self, docs: Iterable[Doc]) -> torch.Tensor:
        self.require_model()
        x = get_last_hidden_state_from_docs(docs)
        return self.model(x)

    @staticmethod
    def install_extensions():
        Doc.set_extension("tokens_logit", default=None, force=True)


TrfForTokenClassificationBase.install_extensions()
TRANSFORMERS_NER = "transformers_ner"


@spacy.component(
    TRANSFORMERS_NER,
    requires=[f"doc._.{ATTRS.last_hidden_state}"],
    assigns=["doc.ents"],
)
class TrfForNamedEntityRecognition(TrfForTokenClassificationBase):
    """Named entity recognition component with pytorch-transformers."""

    K_BEAM = "k_beam"
    DEFAULT_BEAM_WIDTH = 5

    @property
    def ignore_label_index(self) -> int:
        if UNK in self.labels:
            return self.labels.index(UNK)
        return -1

    def compute_loss(
        self, docs: Sequence[Doc], golds: Sequence[GoldParse], outputs: torch.Tensor
    ) -> None:
        target = self._create_target_from_docs_golds(docs, golds, outputs)
        loss = F.cross_entropy(
            outputs.transpose(1, 2), target, ignore_index=self.ignore_label_index
        )
        add_loss_to_docs(docs, loss)

    def set_annotations(
        self, docs: Iterable[Doc], logits: torch.Tensor
    ) -> Iterable[Doc]:
        assert len(logits.shape) == 3  # (batch, length, nclass)
        id2label = self.labels

        for doc, logit in zip(docs, cast(Iterable, logits)):
            doc._.set("tokens_logit", logit)
            best_tags = get_best_tags(logit, id2label, self.k_beam)
            ents = [best_tags[a[0]] if len(a) else "O" for a in doc._.get(ATTRS.align)]
            biluo_ents = iob_to_biluo(ents)
            doc.ents = tuple(
                spacy.util.filter_spans(
                    doc.ents + tuple(spans_from_biluo_tags(doc, biluo_ents))
                )
            )
        return docs

    def _create_target_from_docs_golds(
        self, docs: Sequence[Doc], golds: Sequence[GoldParse], logits: torch.Tensor
    ) -> torch.Tensor:
        all_aligns = (doc._.get(ATTRS.align) for doc in docs)
        new_ners = (
            _convert_goldner(gold.ner or [], align)
            for gold, align in zip(golds, all_aligns)
        )
        return _create_target(new_ners, logits, self.ignore_label_index, self.label2id)

    def _get_nerlabel_from_gold(self, gold: GoldParse) -> Iterator[int]:
        if gold.ner is not None:
            for ner in gold.ner:
                yield self.label2id[self.convert_label(ner)]

    def _extract_logit(
        self, logit: torch.Tensor, alignment: List[List[int]]
    ) -> torch.Tensor:
        idx = [a[0] if len(a) > 0 else self.ignore_label_index for a in alignment]
        return logit[idx]

    @property
    def k_beam(self) -> int:
        return self.cfg.setdefault(self.K_BEAM, self.DEFAULT_BEAM_WIDTH)

    @k_beam.setter
    def k_beam(self, k: int):
        assert isinstance(k, int)
        self.cfg[self.K_BEAM] = k


def _convert_goldner(
    labels: Iterable[str], alignment: List[List[int]]
) -> Dict[int, str]:
    new_ner = {}
    prefixmap = {L: I, U: B}
    for label, align in zip(cast(Iterable[str], labels), alignment):
        prefix, type_ = deconstruct_biluo_label(label)
        new_prefix = prefixmap.get(prefix, prefix)
        for i in align:
            new_ner[i] = construct_biluo_tag(new_prefix, type_)
            if new_prefix == B:
                new_prefix = I
    return new_ner


def _create_target(
    new_ners: Iterable[Dict[int, str]],
    logits: torch.Tensor,
    ignore_index: int,
    label2id: Dict[str, int],
) -> torch.Tensor:
    batch_size, length, _ = logits.shape
    target = logits.new_full((batch_size, length), ignore_index, dtype=torch.long)
    for i, ners in enumerate(new_ners):
        idx_ners = [(i, label2id[ner]) for i, ner in ners.items()]
        if not idx_ners:
            continue
        idx, _ners = zip(*idx_ners)
        target[i, idx] = target.new_tensor(_ners)
    return target


def get_best_tags(logit: torch.Tensor, id2label: List[str], k_beam: int) -> List[str]:
    """Select best tags from logit based on beamsearch."""
    if k_beam > 1:
        logit = logit.softmax(-1)
    candidates = beamsearch(logit, k_beam)
    assert len(cast(Sized, candidates))
    best_tags: List[str] = []
    for cand in candidates:
        tags, is_correct = correct_bio_tags([id2label[j] for j in cast(Iterable, cand)])
        if is_correct:
            best_tags = tags
            break
        if not best_tags:
            best_tags = tags
    best_tags = [t if t != "-" else "O" for t in best_tags]
    return best_tags



# flake8: noqa
from .pattern_search import PatternSearcher
from .regex_ruler import MultipleRegexRuler, RegexRuler
from .transformers import (
    TrfForMultiLabelSequenceClassification,
    TrfForNamedEntityRecognition,
    TrfForSequenceClassification,
    TrfModel,
    TrfTokenizer,
)



import copy
import warnings
from itertools import chain
from typing import Callable, Iterable, List, Sequence, Tuple, TypeVar, Union, overload

import numpy as np
import torch
from spacy.gold import iob_to_biluo
from spacy.tokens import Doc, Span, Token
from spacy.util import filter_spans

from camphr.errors import Warnings


class BILUO:
    B = "B"
    I = "I"  # noqa: E741
    L = "L"
    U = "U"
    O = "O"  # noqa: E741
    UNKNOWN = "-"


B = BILUO.B
I = BILUO.I  # noqa: E741
L = BILUO.L
U = BILUO.U
O = BILUO.O  # noqa: E741
UNK = BILUO.UNKNOWN


def biluo_type(tag: str) -> str:
    for prefix, biluo in [("B-", B), ("I-", I), ("L-", L), ("U-", U)]:
        if tag.startswith(prefix):
            return biluo
    if tag == "O":
        return O
    return UNK


def deconstruct_biluo_label(label: str) -> Tuple[str, str]:
    """Deconstruct biluo label into BILUO prefix and its label type"""
    biluo = biluo_type(label)
    if biluo == UNK:
        return biluo, ""
    if biluo == O:
        return biluo, ""
    return biluo, label[2:]


def is_group(tagl: BILUO, bl: str, tagr: BILUO, br: str) -> bool:
    if bl != br:
        return False
    return tagl in {B, I} and tagr in {I, L}


def construct_biluo_tag(biluo: str, body: str = "") -> str:
    if body:
        assert biluo not in {O, UNK}
        return biluo + "-" + body
    assert biluo in {O, UNK}
    return biluo


def bio_to_biluo(tags: List[str]) -> List[str]:
    warnings.warn("Use spacy.gold.iob_to_biluo instead", DeprecationWarning)
    return iob_to_biluo(tags)


def biluo_to_bio(tags: List[str]) -> List[str]:
    """convert biluo tags to bio tags. Input `tags` is expected to be syntactically correct."""
    tags = copy.copy(tags)
    for i, tag in enumerate(tags):
        t, b = deconstruct_biluo_label(tag)
        if t == L:
            tags[i] = construct_biluo_tag(I, b)
        elif t == U:
            tags[i] = construct_biluo_tag(B, b)
    return tags


def correct_biluo_tags(tags: List[str]) -> Tuple[List[str], bool]:
    """Check and correct biluo tags list so that it can be assigned to `spacy.gold.spans_from_biluo_tags`.

    All invalid tags will be replaced with `-`
    """
    is_correct = True
    tags = ["O"] + copy.copy(tags) + ["O"]
    for i in range(len(tags) - 1):
        tagl = tags[i]
        tagr = tags[i + 1]

        type_l, body_l = deconstruct_biluo_label(tagl)
        type_r, body_r = deconstruct_biluo_label(tagr)

        # left check
        if type_l in {B, I} and not ((type_r == I or type_r == L) and body_l == body_r):
            is_correct = False
            tags[i] = UNK

        # right check
        if type_r in {I, L} and not ((type_l == B or type_l == I) and body_l == body_r):
            is_correct = False
            tags[i + 1] = UNK
    return tags[1:-1], is_correct


def correct_bio_tags(tags: List[str]) -> Tuple[List[str], bool]:
    """Check and correct bio tags list.

    All invalid tags will be replaced with `-`
    """
    tags = copy.copy(tags)
    is_correct = True
    for i, (tagl, tagr) in enumerate(zip(tags, tags[1:])):
        tl, bl = deconstruct_biluo_label(tagl)
        tr, br = deconstruct_biluo_label(tagr)
        # Convert invalid I to B
        if tr == I and not (tl in {B, I} and bl == br):
            tags[i + 1] = construct_biluo_tag(B, br)
    return tags, is_correct


def merge_entities(ents0: Iterable[Span], ents1: Iterable[Span]) -> List[Span]:
    """Merge two ents. ents1 is prior to ents0"""
    Warnings.W0("merge_entities", "spacy.util.filter_spans")
    return filter_spans(list(ents1) + list(ents0))


@overload
def set_heads(doc: Span, heads: List[int]) -> Span:
    ...


@overload
def set_heads(doc: Doc, heads: List[int]) -> Doc:
    ...


def set_heads(doc, heads):
    """Set heads to doc in UD annotation style.

    If fail to set, return doc without doing anything.
    """
    if max(heads) > len(doc) or min(heads) < 0:
        return doc
    for head, token in zip(heads, doc):
        if head == 0:
            token.head = token
        else:
            token.head = doc[head - 1]
    return doc


def get_doc_vector_via_tensor(doc: Doc) -> np.ndarray:
    return doc.tensor.sum(0)


def get_span_vector_via_tensor(span: Span) -> np.ndarray:
    return span.doc.tensor[span.start : span.end].sum(0)


def get_token_vector_via_tensor(token: Token) -> np.ndarray:
    return token.doc.tensor[token.i]


def get_similarity(o1: Union[Doc, Span, Token], o2: Union[Doc, Span, Token]) -> float:
    v1: np.ndarray = o1.vector
    v2: np.ndarray = o2.vector
    return (v1.dot(v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))).item()


USER_HOOKS = "user_hooks"


class UserHooksMixin:
    @property
    def user_hooks(self):
        return self.cfg.setdefault(USER_HOOKS, {})  # type: ignore

    def add_user_hook(self, k: str, fn: Callable):
        hooks = self.user_hooks
        hooks[k] = fn


def beamsearch(probs: torch.Tensor, k: int) -> torch.Tensor:
    """Beam search for sequential probabilities.

    Args:
        probs: tensor of shape (length, d). requires d > 0. Assumed all items in `probs` in range [0, 1].
        k: beam width
    Returns: (k, length) tensor
    """
    assert len(probs.shape) == 2
    if len(probs) == 0:
        return torch.zeros(k, 0)
    if k == 1:
        return probs.argmax(-1)[None, :]

    # We calculate top k-th argmax of E = p0p1p2..pn-1.
    # To avoid over(under)flow, evaluete log(E) instead of E.
    # By doing this, E can be evaluated by addition.
    data = probs.to(torch.double).log()
    _, m = data.shape
    scores, candidates = torch.topk(data[0], k=min(k, m))
    candidates = candidates[:, None]
    for row in data[1:]:
        z = (scores[:, None] + row[None, :]).flatten()
        scores, flat_idx = torch.topk(z, k=min(k, len(z)))
        i, j = flat_idx // m, flat_idx % m
        candidates = torch.cat([candidates[i], j[:, None]], dim=-1)
    return candidates


def flatten_docs_to_sents(docs: Iterable[Doc]) -> List[Span]:
    return list(chain.from_iterable(list(doc.sents) for doc in docs))


T = TypeVar("T")


def chunk(seq: Sequence[T], nums: Sequence[int]) -> List[Sequence[T]]:
    i = 0
    output = []
    for n in nums:
        j = i + n
        output.append(seq[i:j])
        i = j
    return output



from pathlib import Path

import fire
import yaml

MODEL_CONFIG_PATH = Path(__file__).parent / "../model_config"


def main(name: str = "", ls: bool = False):
    if ls:
        for fpath in MODEL_CONFIG_PATH.glob("**/*.yml"):
            with fpath.open() as f:
                cfg = yaml.safe_load(f)
                print(cfg["name"])
        return
    fpath = MODEL_CONFIG_PATH / f"{name}.yml"
    print(fpath.read_text())


if __name__ == "__main__":
    fire.Fire(main)



import logging
import os
import random
from collections import defaultdict
from pathlib import Path
from typing import Callable, Dict, Tuple, Type, Union, cast

import hydra
import hydra.utils
import numpy as np
import torch
from omegaconf import Config, OmegaConf
from sklearn.metrics import classification_report  # type: ignore
from spacy.language import Language
from spacy.util import minibatch
from torch.optim.optimizer import Optimizer

from camphr.cli.utils import (
    InputData,
    check_nonempty,
    convert_fullpath_if_path,
    create_data,
    report_fail,
    unzip2,
)
from camphr.lang.torch import TorchLanguage
from camphr.models import correct_model_config, create_model
from camphr.pipelines.transformers.seq_classification import TOP_LABEL
from camphr.torch_utils import goldcat_to_label
from camphr.utils import (
    create_dict_from_dotkey,
    get_by_dotkey,
    import_attr,
    resolve_alias,
)

logger = logging.getLogger(__name__)


MUST_FIELDS = [
    "train.data.path",
    [
        "model.ner_label",
        "model.textcat_label",
        "model.multitextcat_label",
        "model.pipeline.transformers_ner.labels",
        "model.pipeline.transformers_seq_classification.labels",
        "model.pipeline.transformers_multilabel_seq_classification.labels",
        "model.labels",
        "model.task",
    ],
    "model.lang.name",
]

PATH_FIELDS = [
    "model.ner_label",
    "model.textcat_label",
    "model.multitextcat_label",
    "model.labels",
    "train.data.path",
    "model.pretrained",
]

ALIASES = {
    "train.optimizer": "model.lang.optimizer",
    "data": "train.data",
    "lang": "model.lang.name",
}


def resolve_path(cfg: Config) -> Config:
    for key in PATH_FIELDS:
        path = get_by_dotkey(cfg, key)
        if path:
            path = convert_fullpath_if_path(path)
            cfg = OmegaConf.merge(
                cfg, OmegaConf.create(create_dict_from_dotkey(key, path))
            )
    return cfg


def parse(cfg: Config):
    assert isinstance(cfg, Config), cfg
    cfg = resolve_alias(ALIASES, cfg)
    check_nonempty(cfg, MUST_FIELDS)
    cfg = resolve_path(cfg)
    cfg.model = correct_model_config(cfg.model)
    return cfg


def evaluate_textcat(cfg: Config, nlp: TorchLanguage, val_data: InputData) -> Dict:
    # TODO: https://github.com/explosion/spaCy/pull/4664
    texts, golds = cast(Tuple[Tuple[str], Dict], zip(*val_data))
    try:
        y = np.array(list(map(lambda x: goldcat_to_label(x["cats"]), golds)))
        docs = list(nlp.pipe(texts, batch_size=cfg.nbatch * 2))
        preds = np.array([doc._.get(TOP_LABEL) for doc in docs])
    except Exception:
        report_fail(val_data)
        raise
    return classification_report(y, preds, output_dict=True)


def evaluate(cfg: Config, nlp: TorchLanguage, val_data: InputData) -> Dict:
    try:
        scores = nlp.evaluate(val_data, batch_size=cfg.nbatch * 2)
    except Exception:
        report_fail(val_data)
        raise
    return scores


EvalFn = Callable[[Config, TorchLanguage, InputData], Dict]

EVAL_FN_MAP = defaultdict(  # type: ignore
    lambda: evaluate, {"textcat": evaluate_textcat}  # type: ignore
)


def train_epoch(
    cfg: Config,
    nlp: TorchLanguage,
    optim: Optimizer,
    train_data: InputData,
    val_data: InputData,
    epoch: int,
    eval_fn: EvalFn,
) -> None:
    for j, batch in enumerate(minibatch(train_data, size=cfg.nbatch)):
        texts, golds = unzip2(batch)
        try:
            nlp.update(texts, golds, optim, verbose=True)
        except Exception:
            report_fail(batch)
            raise
        logger.info(f"epoch {epoch} {j*cfg.nbatch}/{cfg.data.ndata}")


def save_model(nlp: Language, path: Path) -> None:
    nlp.to_disk(path)
    logger.info(f"Saved the model in {str(path.absolute())}")


class DummyScheduler:
    @staticmethod
    def step() -> None:
        ...


def load_scheduler(
    cfg: Config, optimizer: Optimizer
) -> Union[torch.optim.lr_scheduler.LambdaLR, Type[DummyScheduler]]:
    cls_str = get_by_dotkey(cfg, "scheduler.class")
    if not cls_str:
        return DummyScheduler
    cls = cast(Type[torch.optim.lr_scheduler.LambdaLR], import_attr(cls_str))
    params = OmegaConf.to_container(cfg.scheduler.params) or {}
    return cls(optimizer, **params)


def train(
    cfg: Config,
    nlp: TorchLanguage,
    train_data: InputData,
    val_data: InputData,
    savedir: Path,
) -> None:
    eval_fn = EVAL_FN_MAP[cfg.task]
    optim = nlp.resume_training()
    scheduler = load_scheduler(cfg, optim)
    for i in range(cfg.niter):
        random.shuffle(train_data)
        train_epoch(cfg, nlp, optim, train_data, val_data, i, eval_fn)
        scheduler.step()  # type: ignore # (https://github.com/pytorch/pytorch/pull/26531)
        scores = eval_fn(cfg, nlp, val_data)
        nlp.meta.update({"score": scores, "config": OmegaConf.to_container(cfg)})
        save_model(nlp, savedir / str(i))


def set_seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)  # type: ignore


def validate_data(cfg: Config, data: InputData, n_check=100):
    if "textcat" in cfg.model.pipeline:
        data = random.sample(data, min(n_check, len(data)))
        for text, gold in data:
            assert "cats" in gold, "`cats` key is required in gold label"
            assert (
                abs(sum(gold["cats"].values()) - 1) < 1e-2
            ), "Sum of gold.cats must equal 1. for text classification task."


def _main(cfg: Config) -> None:
    if cfg.user_config is not None:
        # Override config by user config.
        # This `user_config` have some limitations, and it will be improved
        # after the issue https://github.com/facebookresearch/hydra/issues/386 solved
        cfg = OmegaConf.merge(
            cfg, OmegaConf.load(hydra.utils.to_absolute_path(cfg.user_config))
        )
    cfg = parse(cfg)
    if cfg.seed:
        set_seed(cfg.seed)
    logger.info(cfg.pretty())
    nlp = cast(TorchLanguage, create_model(cfg.model))
    train_data, val_data = create_data(cfg.train.data)
    validate_data(cfg, train_data)
    logger.info("output dir: {}".format(os.getcwd()))
    if torch.cuda.is_available():
        logger.info("CUDA enabled")
        nlp.to(torch.device("cuda"))
    savedir = Path.cwd() / "models"
    savedir.mkdir(exist_ok=True)
    train(cfg.train, nlp, train_data, val_data, savedir)


# Avoid to use decorator for testing
main = hydra.main(config_path="conf/train/config.yaml", strict=False)(_main)

if __name__ == "__main__":
    main()



"""Pipeline of Embed Rank (https://arxiv.org/pdf/1801.04470.pdf).

`EmbedRank` assigns `Doc` extension `embedrank_keyphrases`
"""
from typing import Callable, Dict, List, Optional

import numpy as np
import spacy
import spacy.language
from sklearn.metrics.pairwise import cosine_similarity
from spacy.matcher import Matcher
from spacy.tokens import Doc, Span
from spacy.vocab import Vocab

from camphr.utils import SerializationMixin

spacy.language.ENABLE_PIPELINE_ANALYSIS = True


EMBEDRANK_KEYPHRASES = "embedrank_keyphrases"


class ExtractKeywordsRuler:
    def __init__(self, vocab: Vocab, patterns: Dict[str, List]):
        self.matcher = Matcher(vocab)
        for k, v in patterns.items():
            self.matcher.add(k, None, v)

    def __call__(self, doc: Doc) -> List[Span]:
        return [doc[i:j] for _, i, j in self.matcher(doc)]


@spacy.component(
    "embedrank",
    assigns=[f"doc._.{EMBEDRANK_KEYPHRASES}"],
    requires=["doc.vector", "token.vector"],
)
class EmbedRank(SerializationMixin):
    DefaultPatterns = {
        "keyword": [
            {"TAG": {"REGEX": "(名詞|形容詞|助詞,連体化|接頭詞,名詞接続).*"}, "OP": "*"},
            {"TAG": {"REGEX": "名詞,.*"}, "OP": "+"},
        ]
    }

    serialization_fields = ["max_keyphrases", "extract_keyphrases", "lambda_"]

    @staticmethod
    def install_extensions():
        Doc.set_extension(EMBEDRANK_KEYPHRASES, default=None)

    def __init__(
        self,
        vocab: Optional[Vocab] = None,
        max_keyphrases: int = -1,
        extract_keyphrases_fn: Optional[Callable[[Doc], List[Span]]] = None,
        lambda_: float = 0.5,
    ):
        """

        Args:
            max_keyphrases: max number of keyphrases set in doc._.embedrank_keyphrases. If -1, all keyphrases are set.
            extract_keyphrases_fn: Function that returns the candidates of keyword.
            lambda_: Hyperparameter of Maximal Marginal Relevance (MMR). See the paper (https://arxiv.org/pdf/1801.04470.pdf) for details.
        """
        self.vocab = vocab
        self.max_keyphrases = max_keyphrases
        self.lambda_ = lambda_
        self.extract_keyphrases = extract_keyphrases_fn

    @classmethod
    def from_nlp(cls, nlp: spacy.language.Language) -> "EmbedRank":
        return cls(nlp.vocab)

    def require_model(self):
        if self.extract_keyphrases is None:
            assert self.vocab is not None
            self.extract_keyphrases = ExtractKeywordsRuler(
                self.vocab, self.DefaultPatterns
            )

    def __call__(self, doc: Doc) -> Doc:
        """Extract keyphrases from doc.vector and span.vector, and set them into Doc._.embed_keyphrases sorted by score."""
        self.require_model()
        spans: List[Span] = self.extract_keyphrases(doc)  # type: ignore
        # we know `extract_keyphrases` cannot be `None` after `require_model`
        if not spans:
            return doc
        spans_vectors = np.array([np.array(span.vector) for span in spans])
        spans_sims = cosine_similarity(spans_vectors, spans_vectors)
        scores0 = self.lambda_ * cosine_similarity(
            spans_vectors, doc.vector.reshape(1, -1)
        )
        scores0 = np.squeeze(scores0)

        candidates = list(range(len(spans)))
        i = np.argmax(scores0)
        candidates.remove(i)
        populated: List[int] = [i]

        while candidates and (
            self.max_keyphrases <= 0 or len(populated) < self.max_keyphrases
        ):
            scores = scores0[candidates] - (1 - self.lambda_) * np.max(
                spans_sims[candidates][:, populated], axis=1
            )
            i = np.argmax(scores)
            populated.append(candidates[i])
            candidates.pop(i)
        doc._.set(EMBEDRANK_KEYPHRASES, [spans[i] for i in populated])
        return doc


EmbedRank.install_extensions()
